@@ -3,6 +3,7 @@
 #include <fcitx/candidate.h>
 #include <fcitx/hook.h>
 #include <fcitx-config/xdg.h>
+#include <fcitx/module/freedesktop-notify/fcitx-freedesktop-notify.h>
 #include <libintl.h>
 #include <rime_api.h>
 
@@ -11,6 +12,10 @@
 typedef struct _FcitxRime {
     FcitxInstance* owner;
     RimeSessionId session_id;
+    char* iconname;
+    RimeApi* api;
+    boolean firstRun;
+    FcitxUIMenu schemamenu;
 } FcitxRime;
 
 static void* FcitxRimeCreate(FcitxInstance* instance);
@@ -23,11 +28,15 @@ static INPUT_RETURN_VALUE FcitxRimeDoReleaseInput(void* arg, FcitxKeySym sym, un
 static INPUT_RETURN_VALUE FcitxRimeDoInputReal(void* arg, FcitxKeySym _sym, unsigned int _state);
 static INPUT_RETURN_VALUE FcitxRimeGetCandWords(void* arg);
 static void FcitxRimeToggleEnZh(void* arg);
-static const char* FcitxRimeGetDummy(void* arg);
+static const char* FcitxRimeGetIMIcon(void* arg);
+static const char* FcitxRimeGetDeployIcon(void *arg);
+static const char* FcitxRimeGetSyncIcon(void *arg);
 static void FcitxRimeToggleSync(void* arg);
 static void FcitxRimeToggleDeploy(void* arg);
 static void FcitxRimeResetUI(void* arg);
 static void FcitxRimeUpdateStatus(FcitxRime* rime);
+static boolean FcitxRimeSchemaMenuAction(FcitxUIMenu *menu, int index);
+static void FcitxRimeSchemaMenuUpdate(FcitxUIMenu *menu);
 
 FCITX_EXPORT_API
 FcitxIMClass ime = {
@@ -38,6 +47,29 @@ FcitxIMClass ime = {
 FCITX_EXPORT_API
 int ABI_VERSION = FCITX_ABI_VERSION;
 
+void FcitxRimeNotificationHandler(void* context_object,
+                                  RimeSessionId session_id,
+                                  const char* message_type,
+                                  const char* message_value)
+{
+    const char* message = NULL;
+    FcitxRime* rime = (FcitxRime*) context_object;
+    if (!strcmp(message_type, "deploy")) {
+        if (!strcmp(message_value, "start")) {
+            message = _("Rime is under maintenance ...");
+        } else if (!strcmp(message_value, "success")) {
+            message = _("Rime is ready.");
+        } else if (!strcmp(message_value, "failure")) {
+            message = _("Rime has encountered an error. "
+                        "See /tmp/rime.fcitx.ERROR for details.");
+        }
+    }
+
+    if (message) {
+        FcitxFreeDesktopNotifyShowAddonTip(rime->owner, "fcitx-rime-deploy", "fcitx-rime-deploy", _("Rime"), message);
+    }
+}
+
 static void FcitxRimeStart(FcitxRime* rime, boolean fullcheck) {
 
     char* user_path = NULL;
@@ -48,24 +80,35 @@ static void FcitxRimeStart(FcitxRime* rime, boolean fullcheck) {
     //char* shared_data_dir = fcitx_utils_get_fcitx_path_with_filename("pkgdatadir", "rime");
     const char* shared_data_dir = RIME_DATA_DIR;
 
-    RimeTraits fcitx_rime_traits;
+    RIME_STRUCT(RimeTraits, fcitx_rime_traits);
     fcitx_rime_traits.shared_data_dir = shared_data_dir;
+    fcitx_rime_traits.app_name = "rime.fcitx-rime";
     fcitx_rime_traits.user_data_dir = user_path;
     fcitx_rime_traits.distribution_name = "Rime";
     fcitx_rime_traits.distribution_code_name = "fcitx-rime";
-    fcitx_rime_traits.distribution_version = "0.1";
-    RimeInitialize(&fcitx_rime_traits);
-    if (RimeStartMaintenance(fullcheck)) {
-        // TODO: notification...
+    fcitx_rime_traits.distribution_version = "0.2.3";
+    if (rime->firstRun) {
+        rime->api->setup(&fcitx_rime_traits);
+        rime->firstRun = false;
     }
+    rime->api->initialize(&fcitx_rime_traits);
+    rime->api->set_notification_handler(FcitxRimeNotificationHandler, rime);
+    rime->api->start_maintenance(fullcheck);
 
-    rime->session_id = RimeCreateSession();
+    rime->session_id = rime->api->create_session();
 }
 
 static void* FcitxRimeCreate(FcitxInstance* instance)
 {
     FcitxRime* rime = (FcitxRime*) fcitx_utils_malloc0(sizeof(FcitxRime));
     rime->owner = instance;
+    rime->api = rime_get_api();
+    rime->firstRun = true;
+    if (!rime->api) {
+        free(rime);
+        return NULL;
+    }
+
     FcitxRimeStart(rime, false);
 
     FcitxIMIFace iface;
@@ -91,38 +134,47 @@ static void* FcitxRimeCreate(FcitxInstance* instance)
     FcitxUIRegisterComplexStatus(
         instance,
         rime,
-        "rimeenzh",
+        "rime-enzh",
         "",
         "",
         FcitxRimeToggleEnZh,
-        FcitxRimeGetDummy);
+        FcitxRimeGetIMIcon);
 
     FcitxUIRegisterComplexStatus(
         instance,
         rime,
-        "rimedeploy",
-        _("\xe2\x9f\xb2 Deploy"),
+        "rime-deploy",
+        _("Deploy"),
         _("Deploy"),
         FcitxRimeToggleDeploy,
-        FcitxRimeGetDummy);
+        FcitxRimeGetDeployIcon);
 
     FcitxUIRegisterComplexStatus(
         instance,
         rime,
-        "rimesync",
-        _("\xe2\x87\x85 Synchronize"),
+        "rime-sync",
+        _("Synchronize"),
         _("Synchronize"),
         FcitxRimeToggleSync,
-        FcitxRimeGetDummy);
+        FcitxRimeGetSyncIcon);
 
-    FcitxUISetStatusVisable(instance, "rimeenzh", false);
-    FcitxUISetStatusVisable(instance, "rimesync", false);
-    FcitxUISetStatusVisable(instance, "rimedeploy", false);
+    FcitxUISetStatusVisable(instance, "rime-enzh", false);
+    FcitxUISetStatusVisable(instance, "rime-sync", false);
+    FcitxUISetStatusVisable(instance, "rime-deploy", false);
     FcitxIMEventHook hk;
     hk.arg = rime;
     hk.func = FcitxRimeResetUI;
 
     FcitxInstanceRegisterResetInputHook(instance, hk);
+    
+    FcitxMenuInit(&rime->schemamenu);
+    rime->schemamenu.name = strdup(_("Schema List"));
+    rime->schemamenu.candStatusBind = strdup("rime-enzh");
+    rime->schemamenu.MenuAction = FcitxRimeSchemaMenuAction;
+    rime->schemamenu.UpdateMenu = FcitxRimeSchemaMenuUpdate;
+    rime->schemamenu.priv = rime;
+    rime->schemamenu.isSubMenu = false;
+    FcitxUIRegisterMenu(rime->owner, &rime->schemamenu);
 
     return rime;
 }
@@ -131,10 +183,16 @@ void FcitxRimeDestroy(void* arg)
 {
     FcitxRime* rime = (FcitxRime*) arg;
     if (rime->session_id) {
-        RimeDestroySession(rime->session_id);
+        rime->api->destroy_session(rime->session_id);
         rime->session_id = 0;
     }
-    RimeFinalize();
+
+    FcitxUIUnRegisterMenu(rime->owner, &rime->schemamenu);
+    FcitxMenuFinalize(&rime->schemamenu);
+    
+    fcitx_utils_free(rime->iconname);
+    rime->api->finalize();
+    free(rime);
 }
 
 boolean FcitxRimeInit(void* arg)
@@ -155,11 +213,15 @@ boolean FcitxRimeInit(void* arg)
 void FcitxRimeReset(void* arg)
 {
     FcitxRime* rime = (FcitxRime*) arg;
-    if (!RimeFindSession(rime->session_id)) {
-        rime->session_id = RimeCreateSession();
+
+    if (rime->api->is_maintenance_mode()) {
+        return;
+    }
+    if (!rime->api->find_session(rime->session_id)) {
+        rime->session_id = rime->api->create_session();
     }
     if (rime->session_id) {
-        RimeProcessKey(rime->session_id, FcitxKey_Escape, 0);
+        rime->api->process_key(rime->session_id, FcitxKey_Escape, 0);
     }
 }
 
@@ -182,9 +244,15 @@ INPUT_RETURN_VALUE FcitxRimeDoReleaseInput(void* arg, FcitxKeySym _sym, unsigned
 
 void FcitxRimeUpdateStatus(FcitxRime* rime)
 {
-    RimeStatus status = {0};
-    RIME_STRUCT_INIT(RimeStatus, status);
-    if (RimeGetStatus(rime->session_id, &status)) {
+    if (rime->api->is_maintenance_mode()) {
+        return;
+    }
+    if (!rime->api->find_session(rime->session_id)) {
+        rime->session_id = rime->api->create_session();
+    }
+    
+    RIME_STRUCT(RimeStatus, status);
+    if (rime->api->get_status(rime->session_id, &status)) {
         char* text = "";
         if (status.is_disabled) {
             text = "\xe2\x8c\x9b";
@@ -196,8 +264,10 @@ void FcitxRimeUpdateStatus(FcitxRime* rime)
         } else {
             text = "中";
         }
-        FcitxUISetStatusString(rime->owner, "rimeenzh", text, text);
-        RimeFreeStatus(&status);
+        FcitxUISetStatusString(rime->owner, "rime-enzh", text, text);
+        rime->api->free_status(&status);
+    } else {
+        FcitxUISetStatusString(rime->owner, "rime-enzh", "\xe2\x8c\x9b", "\xe2\x8c\x9b");
     }
 }
 
@@ -207,13 +277,13 @@ INPUT_RETURN_VALUE FcitxRimeDoInput(void* arg, FcitxKeySym _sym, unsigned int _s
     FcitxInputState *input = FcitxInstanceGetInputState(rime->owner);
     uint32_t sym = FcitxInputStateGetKeySym(input);
     uint32_t state = FcitxInputStateGetKeyState(input);
-    _state &= (FcitxKeyState_SimpleMask);
+    _state &= (FcitxKeyState_SimpleMask | FcitxKeyState_CapsLock);
 
-    if (_state & (~(FcitxKeyState_Ctrl_Alt_Shift))) {
+    if (_state & (~(FcitxKeyState_Ctrl_Alt_Shift | FcitxKeyState_CapsLock))) {
         return IRV_TO_PROCESS;
     }
 
-    state &= (FcitxKeyState_SimpleMask);
+    state &= (FcitxKeyState_SimpleMask | FcitxKeyState_CapsLock);
 
     return FcitxRimeDoInputReal(arg, sym, state);
 }
@@ -222,19 +292,23 @@ INPUT_RETURN_VALUE FcitxRimeDoInputReal(void* arg, FcitxKeySym sym, unsigned int
 {
     FcitxRime *rime = (FcitxRime *)arg;
 
-    if (!RimeFindSession(rime->session_id)) {
-        rime->session_id = RimeCreateSession();
-    }
-    if (!rime->session_id) { // service disabled
+    if (rime->api->is_maintenance_mode()) {
         return IRV_TO_PROCESS;
     }
-    boolean result = RimeProcessKey(rime->session_id, sym, state);
+    if (!rime->api->find_session(rime->session_id)) {
+        rime->session_id = rime->api->create_session();
+    }
+    if (!rime->session_id) { // service disabled
+        FcitxRimeUpdateStatus(rime);
+        return IRV_TO_PROCESS;
+    }
+    boolean result = rime->api->process_key(rime->session_id, sym, state);
 
-    RimeCommit commit = {0};
-    if (RimeGetCommit(rime->session_id, &commit)) {
+    RIME_STRUCT(RimeCommit, commit);
+    if (rime->api->get_commit(rime->session_id, &commit)) {
         FcitxInputContext* ic = FcitxInstanceGetCurrentIC(rime->owner);
         FcitxInstanceCommitString(rime->owner, ic, commit.text);
-        RimeFreeCommit(&commit);
+        rime->api->free_commit(&commit);
     }
 
     FcitxRimeUpdateStatus(rime);
@@ -250,16 +324,15 @@ INPUT_RETURN_VALUE FcitxRimeDoInputReal(void* arg, FcitxKeySym sym, unsigned int
 
 INPUT_RETURN_VALUE FcitxRimeGetCandWord(void* arg, FcitxCandidateWord* candWord)
 {
-    RimeContext context = {0};
     FcitxRime *rime = (FcitxRime *)arg;
-    RIME_STRUCT_INIT(RimeContext, context);
+    RIME_STRUCT(RimeContext, context);
     INPUT_RETURN_VALUE retVal = IRV_TO_PROCESS;
-    if (RimeGetContext(rime->session_id, &context)) {
+    if (rime->api->get_context(rime->session_id, &context)) {
         if (context.menu.num_candidates)
         {
             int i = *(int*) candWord->priv;
             const char* digit = DIGIT_STR_CHOOSE;
-            int num_select_keys = strlen(context.menu.select_keys);
+            int num_select_keys = context.menu.select_keys ? strlen(context.menu.select_keys) : 0;
             FcitxKeySym sym = FcitxKey_None;
             if (i < 10) {
                 if (i < num_select_keys)
@@ -268,12 +341,13 @@ INPUT_RETURN_VALUE FcitxRimeGetCandWord(void* arg, FcitxCandidateWord* candWord)
                     sym = digit[i];
             }
             if (sym != FcitxKey_None) {
-                boolean result = RimeProcessKey(rime->session_id, sym, 0);
-                RimeCommit commit = {0};
-                if (RimeGetCommit(rime->session_id, &commit)) {
+                boolean result = rime->api->process_key(rime->session_id, sym, 0);
+                
+                RIME_STRUCT(RimeCommit, commit);
+                if (rime->api->get_commit(rime->session_id, &commit)) {
                     FcitxInputContext* ic = FcitxInstanceGetCurrentIC(rime->owner);
                     FcitxInstanceCommitString(rime->owner, ic, commit.text);
-                    RimeFreeCommit(&commit);
+                    rime->api->free_commit(&commit);
                 }
                 if (!result) {
                     FcitxRimeGetCandWords(rime);
@@ -284,9 +358,9 @@ INPUT_RETURN_VALUE FcitxRimeGetCandWord(void* arg, FcitxCandidateWord* candWord)
                     retVal = IRV_DISPLAY_CANDWORDS;
             }
         }
+        rime->api->free_context(&context);
     }
 
-    RimeFreeContext(&context);
     return retVal;
 }
 
@@ -297,11 +371,13 @@ INPUT_RETURN_VALUE FcitxRimeGetCandWords(void* arg)
     FcitxInputState *input = FcitxInstanceGetInputState(rime->owner);
     FcitxInstanceCleanInputWindow(rime->owner);
 
-    RimeContext context = {0};
-    RIME_STRUCT_INIT(RimeContext, context);
-    if (!RimeGetContext(rime->session_id, &context) ||
-        context.composition.length == 0) {
-        RimeFreeContext(&context);
+    RIME_STRUCT(RimeContext, context);
+    if (!rime->api->get_context(rime->session_id, &context)) {
+        return IRV_DISPLAY_CANDWORDS;
+    }
+
+    if (context.composition.length == 0) {
+        rime->api->free_context(&context);
         return IRV_DISPLAY_CANDWORDS;
     }
 
@@ -347,7 +423,7 @@ INPUT_RETURN_VALUE FcitxRimeGetCandWords(void* arg)
         char strChoose[11];
         strChoose[10] = '\0';
         FcitxCandidateWordSetPageSize(candList, 10);
-        int num_select_keys = strlen(context.menu.select_keys);
+        int num_select_keys = context.menu.select_keys ? strlen(context.menu.select_keys) : 0;
         int i;
         for (i = 0; i < context.menu.num_candidates; ++i) {
             FcitxCandidateWord candWord;
@@ -379,7 +455,7 @@ INPUT_RETURN_VALUE FcitxRimeGetCandWords(void* arg)
         FcitxCandidateWordSetOverridePaging(candList, context.menu.page_no != 0, !context.menu.is_last_page, NULL, NULL, NULL);
     }
 
-    RimeFreeContext(&context);
+    rime->api->free_context(&context);
     return IRV_DISPLAY_CANDWORDS;
 }
 
@@ -387,20 +463,103 @@ void FcitxRimeReloadConfig(void* arg)
 {
     FcitxRime* rime = (FcitxRime*) arg;
     if (rime->session_id) {
-        RimeDestroySession(rime->session_id);
+        rime->api->destroy_session(rime->session_id);
         rime->session_id = 0;
     }
-    RimeFinalize();
+    rime->api->finalize();
     FcitxRimeStart(rime, false);
 
     FcitxRimeUpdateStatus(rime);
 }
 
-const char* FcitxRimeGetDummy(void* arg)
+
+boolean FcitxRimeSchemaMenuAction(FcitxUIMenu* menu, int index)
 {
-    return "";
+    FcitxRime * rime = menu->priv;
+    if (rime->api->is_maintenance_mode()) {
+        return false;
+    }
+    if (!rime->api->find_session(rime->session_id)) {
+        rime->session_id = rime->api->create_session();
+    }
+
+    if (index == 0) {
+        rime->api->set_option(rime->session_id, "ascii_mode", true);
+    } else {
+        rime->api->set_option(rime->session_id, "ascii_mode", false);
+        RimeSchemaList list = {0};
+        if (rime->api->get_schema_list(&list)) {
+            if (list.size > index - 1) {
+                rime->api->select_schema(rime->session_id,list.list[index-1].schema_id);
+                FcitxRimeUpdateStatus(rime);
+                FcitxRimeGetCandWords(rime);
+                FcitxUIUpdateInputWindow(rime->owner);
+            }
+            rime->api->free_schema_list(&list);
+        }
+    }
+
+    return true;
 }
 
+void FcitxRimeSchemaMenuUpdate(FcitxUIMenu* menu)
+{
+    FcitxRime * rime = menu->priv;
+    if (rime->api->is_maintenance_mode()) {
+        return;
+    }
+    if (!rime->api->find_session(rime->session_id)) {
+        rime->session_id = rime->api->create_session();
+    }
+
+    FcitxMenuClear(menu);
+
+    FcitxMenuAddMenuItem(menu, _("English"), MENUTYPE_SIMPLE, NULL);
+    RimeSchemaList list = {0};
+    if (rime->api->get_schema_list(&list)) {
+        size_t i = 0;
+        for (i = 0; i < list.size; ++i) {
+            FcitxMenuAddMenuItem(menu, list.list[i].name, MENUTYPE_SIMPLE, NULL);
+        }
+        rime->api->free_schema_list(&list);
+    }
+}
+
+static const char* FcitxRimeGetIMIcon(void* arg)
+{
+    FcitxRime* rime = (FcitxRime*) arg;
+    RIME_STRUCT(RimeStatus, status);
+    if (rime->api->get_status(rime->session_id, &status)) {
+        char* text = "";
+        if (status.is_disabled) {
+            text = "@rime-disable";
+        } else if (status.is_ascii_mode) {
+            text = "@rime-latin";
+        } else if (status.schema_id) {
+            fcitx_utils_free(rime->iconname);
+            fcitx_utils_alloc_cat_str(rime->iconname, "@rime-im-", status.schema_id);
+            text = rime->iconname;
+        } else {
+            text = "@rime-im";
+        }
+        rime->api->free_status(&status);
+
+        return text;
+    }
+    return "@rime-disable";
+}
+
+static const char* FcitxRimeGetDeployIcon(void *arg)
+{
+    return "rime-deploy";
+}
+
+static const char* FcitxRimeGetSyncIcon(void *arg)
+{
+    return "rime-sync";
+}
+
+
 void FcitxRimeToggleEnZh(void* arg)
 {
 
@@ -417,26 +576,30 @@ void FcitxRimeResetUI(void* arg)
         visible = false;
     else
         visible = true;
-    FcitxUISetStatusVisable(instance, "rimeenzh", visible);
-    FcitxUISetStatusVisable(instance, "rimesync", visible);
-    FcitxUISetStatusVisable(instance, "rimedeploy", visible);
+    FcitxUISetStatusVisable(instance, "rime-enzh", visible);
+    FcitxUISetStatusVisable(instance, "rime-sync", visible);
+    FcitxUISetStatusVisable(instance, "rime-deploy", visible);
 }
 
 void FcitxRimeToggleSync(void* arg)
 {
-    RimeSyncUserData();
+    FcitxRime* rime = (FcitxRime*) arg;
+    rime->api->sync_user_data();
+    FcitxRimeGetCandWords(rime);
+    FcitxUIUpdateInputWindow(rime->owner);
 }
 
 void FcitxRimeToggleDeploy(void* arg)
 {
     FcitxRime* rime = (FcitxRime*) arg;
     if (rime->session_id) {
-        RimeDestroySession(rime->session_id);
+        rime->api->sync_user_data(rime->session_id);
         rime->session_id = 0;
     }
-    RimeFinalize();
+    rime->api->finalize();
     FcitxRimeStart(rime, true);
 
     FcitxRimeUpdateStatus(rime);
+    FcitxRimeGetCandWords(rime);
+    FcitxUIUpdateInputWindow(rime->owner);
 }
-
